library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.FfpgPkg.all;

entity DacsController is
    generic (
        g_ClkFrequency: positive -- input clock frequency in Hz
    );
    port (
        Clk_ik: in std_logic;
        Reset_ir: in std_logic;
        TriggerValue_ib16: in unsigned(15 downto 0);
        TriggerLoad_i: in std_logic;
        VcxoValue_ib16: in unsigned(15 downto 0);
        VcxoLoad_i: in std_logic;
        TriggerDac_o: out t_Ad5600Interface;
        VcxoDac_o: out t_Ad5600Interface;
        TriggerBusy_o: out std_logic;
        VcxoBusy_o: out std_logic
    );
end entity;

architecture syn of DacsController is

    --! computes least possible width of the vector to store value X
    -- http://stackoverflow.com/a/12751341/615627
    function log2(X: positive) return natural is
        variable Result: natural;
    begin
        Result := 0;
        while (2**Result < X) loop
            Result := Result + 1;
        end loop;
        return Result;
    end function;

    function f_calculateSclkDivsel(ClkFrequency, DacFrequencyMax: positive) return std_logic_vector is
        variable Result: integer;
    begin
        Result := log2(ClkFrequency/DacFrequencyMax);
        if Result < 3 then
            Result := 0;
        else
            Result := Result - 3;
        end if;
        assert Result < 8
            report "Clock frequency is too high for use with module gc_serial_dac!"
            severity failure;
        return std_logic_vector(to_unsigned(Result, 3));
    end function;

    constant c_DacFrequencyMax: positive := 20_000_000; -- in Hz
    constant c_SclkDivsel: std_logic_vector(2 downto 0) := f_calculateSclkDivsel(g_ClkFrequency, c_DacFrequencyMax);

    signal Reset_nr: std_logic;
    signal TriggerDacEn, VcxoDacEn: std_logic;
    signal TriggerBusy, VcxoBusy: std_logic;
    
begin

    Reset_nr <= not Reset_ir;

    TriggerDacEn <= TriggerLoad_i and not TriggerBusy;
    VcxoDacEn <= VcxoLoad_i and not VcxoBusy;
    
    cTriggerDac: entity work.gc_serial_dac(syn)
        generic map (
            g_num_data_bits => 16,
            g_num_extra_bits => 8,
            g_num_cs_select => 1,
            g_sclk_polarity => 0
        )
        port map (
            clk_i => Clk_ik,
            rst_n_i => Reset_nr,
            value_i => std_logic_vector(TriggerValue_ib16),
            cs_sel_i(0) => TriggerDacEn,
            load_i => TriggerDacEn,
            sclk_divsel_i => c_SclkDivsel,
            dac_cs_n_o(0) => TriggerDac_o.FrameSynchronization_n,
            dac_sclk_o => TriggerDac_o.SerialClock,
            dac_sdata_o => TriggerDac_o.SerialData,
            busy_o => TriggerBusy
        );

    cVcxoDac: entity work.gc_serial_dac(syn)
        generic map (
            g_num_data_bits => 16,
            g_num_extra_bits => 8,
            g_num_cs_select => 1,
            g_sclk_polarity => 0
        )
        port map (
            clk_i => Clk_ik,
            rst_n_i => Reset_nr,
            value_i => std_logic_vector(VcxoValue_ib16),
            cs_sel_i(0) => VcxoDacEn,
            load_i => VcxoDacEn,
            sclk_divsel_i => c_SclkDivsel,
            dac_cs_n_o(0) => VcxoDac_o.FrameSynchronization_n,
            dac_sclk_o => VcxoDac_o.SerialClock,
            dac_sdata_o => VcxoDac_o.SerialData,
            busy_o => VcxoBusy
        );
    
    TriggerBusy_o <= TriggerBusy;
    VcxoBusy_o <= VcxoBusy;

end architecture;
