library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity DelayController is
    generic (
        g_ClkFrequency: positive -- input clock frequency in Hz
    );
    port (
        Clk_ik: in std_logic;
        Reset_ir: in std_logic;
        Ch1SetValue_ib: in unsigned(9 downto 0);
        Ch1SetValueLoad_i: in std_logic;
        Ch1ResValue_ib: in unsigned(9 downto 0);
        Ch1ResValueLoad_i: in std_logic;
        Ch2SetValue_ib: in unsigned(9 downto 0);
        Ch2SetValueLoad_i: in std_logic;
        Ch2ResValue_ib: in unsigned(9 downto 0);
        Ch2ResValueLoad_i: in std_logic;
        DelayValue_ob: out unsigned(9 downto 0);
        Ch1SetLe_o: out std_logic;
        Ch1ResLe_o: out std_logic;
        Ch2SetLe_o: out std_logic;
        Ch2ResLe_o: out std_logic;
        Busy_o: out std_logic
    );
end entity;

architecture syn of DelayController is

    type t_Output is (e_None, e_Ch1Set, e_Ch1Res, e_Ch2Set, e_Ch2Res);

    signal StartConfiguration: std_logic := '0';
    signal LoadPulse: std_logic := '0';
    signal ActiveOutput: t_Output := e_None;

begin

    pStart: process (Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if Reset_ir = '1' then
                ActiveOutput <= e_None;
                DelayValue_ob <= (others => '0');
                StartConfiguration <= '0';
            elsif LoadPulse = '0' then
                StartConfiguration <= '0';
                if Ch1SetValueLoad_i = '1' then
                    ActiveOutput <= e_Ch1Set;
                    DelayValue_ob <= Ch1SetValue_ib;
                    StartConfiguration <= '1';
                end if;
                if Ch1ResValueLoad_i = '1' then
                    ActiveOutput <= e_Ch1Res;
                    DelayValue_ob <= Ch1ResValue_ib;
                    StartConfiguration <= '1';
                end if;
                if Ch2SetValueLoad_i = '1' then
                    ActiveOutput <= e_Ch2Set;
                    DelayValue_ob <= Ch2SetValue_ib;
                    StartConfiguration <= '1';
                end if;
                if Ch2ResValueLoad_i = '1' then
                    ActiveOutput <= e_Ch2Res;
                    DelayValue_ob <= Ch2ResValue_ib;
                    StartConfiguration <= '1';
                end if;
            end if;
        end if;
    end process;

    cPulseGenerator: entity work.PulseGeneratorTime(syn)
        generic map (
            g_ClkFrequency => g_ClkFrequency,
            g_PulseMinWidthInTime => 1 ns
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Signal_i => StartConfiguration,
            Pulse_o => LoadPulse
        );
    
    Ch1SetLe_o <=
        '0' when (ActiveOutput = e_Ch1Set) and LoadPulse = '1' else
        '1';
    Ch1ResLe_o <=
        '0' when (ActiveOutput = e_Ch1Res) and LoadPulse = '1' else
        '1';
    Ch2SetLe_o <=
        '0' when (ActiveOutput = e_Ch2Set) and LoadPulse = '1' else
        '1';
    Ch2ResLe_o <=
        '0' when (ActiveOutput = e_Ch2Res) and LoadPulse = '1' else
        '1';

    Busy_o <= LoadPulse;

end architecture;
