library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.FfpgPkg.all;

entity Fsm is
    port (
        -- main signals
        Clk_ik: in std_logic;
        Reset_ir: in std_logic;
        -- inputs
        Mode_i: in t_Mode;
        ModeLoad_i: in std_logic;
        Trigger_i: in std_logic;
        StreamReset_i: in std_logic;
        -- outputs
        Running_o: out std_logic := '0';
        GenerationEnable_o: out std_logic := '0';
        OutputEnable_o: out std_logic := '0';
        -- debug
        State_o: out unsigned(2 downto 0)
    );
end entity;

architecture syn of Fsm is

    type t_State is (
        s_Stop,
        s_WaitForTrigger,
        s_Generating,
        s_Outputting
    );
    
    function f_State2Unsigned (State: t_State) return unsigned is begin
        return to_unsigned(t_State'pos(State), 3);
    end function;

    constant c_ResetState: t_State := s_Stop;
    
    signal State: t_State := c_ResetState;

begin

	pFsmTransitions: process (Clk_ik) is begin
		if rising_edge(Clk_ik) then
			if Reset_ir = '1' then
				State <= c_ResetState;
			else
				State <= State;
                if ModeLoad_i = '1' then
                    -- global paths
                    if Mode_i = e_ModeStop then
                        -- stop command
                        State <= s_Stop;
                    else
                        -- start command
                        State <= s_WaitForTrigger;
                    end if;
                else
                    -- separate states
                    case State is
                        when s_WaitForTrigger =>
                            if Trigger_i = '1' then
                                State <= s_Generating;
                            end if;
                        when s_Generating =>
                            if StreamReset_i = '1' then
                                State <= s_Outputting;
                            end if;
                        when s_Outputting =>
                            if (StreamReset_i = '1') and (Mode_i = e_ModeSingle) then
                                State <= s_Stop;
                            end if;
                        when others =>
                            State <= c_ResetState;
                    end case;
                end if;
			end if;
		end if;
	end process;
    
    pFsmOutputs: process (State) is begin
        Running_o <= '0';
        OutputEnable_o <= '0';
        GenerationEnable_o <= '0';
        case State is
            when s_WaitForTrigger =>
                Running_o <= '1';
            when s_Generating =>
                Running_o <= '1';
                GenerationEnable_o <= '1';
            when s_Outputting =>
                Running_o <= '1';
                GenerationEnable_o <= '1';
                OutputEnable_o <= '1';
            when others => null;
        end case;
    end process;
    
    State_o <= f_State2Unsigned(State);

end architecture;