library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.FfpgPkg.all;
use work.ffpg_wbgen2_pkg.all;
use work.wishbone_pkg.all;

entity FfpgSlave is
    generic (
        g_ClkFrequency: positive -- input clock frequency in Hz
    );
    port (
        -- Wishbone connection
        Clk_ik: in std_logic;
        Reset_ir: in std_logic;
        Wb_i: in t_wishbone_slave_in;
        Wb_o: out t_wishbone_slave_out;
        --- FMC interface
        -- clock
        ClkIn0_ik: in std_logic;
        -- DACs
        TriggerDac_o: out t_Ad5600Interface;
        VcxoDac_o: out t_Ad5600Interface;
        -- output enable
        Ch1OutputEnable_o: out std_logic;
        Ch2OutputEnable_o: out std_logic;
        -- delay configuration
        DelayValue_ob: out unsigned(9 downto 0);
        Ch1SetLe_o: out std_logic;
        Ch1ResLe_o: out std_logic;
        Ch2SetLe_o: out std_logic;
        Ch2ResLe_o: out std_logic;
        --
        Ch1Set_o: out std_logic;
        Ch1Res_o: out std_logic;
        Ch2Set_o: out std_logic;
        Ch2Res_o: out std_logic;
        Trigger_i: in std_logic;
        Clk2Sel_o: out std_logic;
        -- LEDs
        Led_ob: out std_logic_vector(4 downto 1);
        -- AD9512 SPI
        SpiAd9512Sclk_o: out std_logic;
        SpiAd9512Mosi_o: out std_logic;
        SpiAd9512Miso_i: in std_logic;
        SpiAd9512Cs_on: out std_logic;
        -- AD9512 func. pin
        Ad9512Func_o: out std_logic;
        -- clock output
        ClkOut_ok: out std_logic
    );
end entity;

architecture syn of FfpgSlave is

    signal ClkRf_k: std_logic;
    signal TriggerRf: std_logic;
    
    signal Ch1SetMemAddress_b11: unsigned(10 downto 0);
    signal Ch1SetMemData_b32: unsigned(31 downto 0);
    signal Ch1SetMemReadStrobe: std_logic;
    signal Ch1ResMemAddress_b11: unsigned(10 downto 0);
    signal Ch1ResMemData_b32: unsigned(31 downto 0);
    signal Ch1ResMemReadStrobe: std_logic;
    signal Ch2SetMemAddress_b11: unsigned(10 downto 0);
    signal Ch2SetMemData_b32: unsigned(31 downto 0);
    signal Ch2SetMemReadStrobe: std_logic;
    signal Ch2ResMemAddress_b11: unsigned(10 downto 0);
    signal Ch2ResMemData_b32: unsigned(31 downto 0);
    signal Ch2ResMemReadStrobe: std_logic;

    signal WbRegsInput: t_ffpg_in_registers;
    signal WbRegsOutput: t_ffpg_out_registers;
    
    signal LedSignal_b: std_logic_vector(4 downto 1);
    
    signal Ad9512ClockSelectionChanged: std_logic;
    signal Ad9512ClockSelection, Ad9512StartConfig: std_logic;
    signal Ad9512SyncePulse: std_logic;
    
begin

    ClkRf_k <= ClkIn0_ik;

    ----------------------------------
    -- Wishbone slave
    ----------------------------------
    cWbSlaveWrapper: entity work.WbSlaveWrapper(syn)
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Wb_i => Wb_i,
            Wb_o => Wb_o,
            ClkRf_ik => ClkRf_k,
            Ch1SetMemAddress_ib11 => Ch1SetMemAddress_b11,
            Ch1SetMemData_ob32 => Ch1SetMemData_b32,
            Ch1SetMemReadStrobe_i => Ch1SetMemReadStrobe,
            Ch1ResMemAddress_ib11 => Ch1ResMemAddress_b11,
            Ch1ResMemData_ob32 => Ch1ResMemData_b32,
            Ch1ResMemReadStrobe_i => Ch1ResMemReadStrobe,
            Ch2SetMemAddress_ib11 => Ch2SetMemAddress_b11,
            Ch2SetMemData_ob32 => Ch2SetMemData_b32,
            Ch2SetMemReadStrobe_i => Ch2SetMemReadStrobe,
            Ch2ResMemAddress_ib11 => Ch2ResMemAddress_b11,
            Ch2ResMemData_ob32 => Ch2ResMemData_b32,
            Ch2ResMemReadStrobe_i => Ch2ResMemReadStrobe,
            WbRegs_i => WbRegsInput,
            WbRegs_o => WbRegsOutput
        );
    
    ----------------------------------
    -- DACs
    ----------------------------------
    cDacsController: entity work.DacsController(syn)
        generic map (
            g_ClkFrequency => g_ClkFrequency
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            TriggerValue_ib16 => WbRegsOutput.trigger_threshold_o,
            TriggerLoad_i => WbRegsOutput.trigger_threshold_load_o,
            VcxoValue_ib16 => WbRegsOutput.vcxo_voltage_o,
            VcxoLoad_i => WbRegsOutput.vcxo_voltage_load_o,
            TriggerDac_o => TriggerDac_o,
            VcxoDac_o => VcxoDac_o,
            TriggerBusy_o => WbRegsInput.status_dac_trigger_busy_i,
            VcxoBusy_o => WbRegsInput.status_dac_vcxo_busy_i
        );

    ----------------------------------
    -- channel delay configuration
    ----------------------------------
    cDelayController: entity work.DelayController(syn)
        generic map (
            g_ClkFrequency => g_ClkFrequency
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Ch1SetValue_ib => WbRegsOutput.ch1_delay_set_o,
            Ch1SetValueLoad_i => WbRegsOutput.ch1_delay_set_load_o,
            Ch1ResValue_ib => WbRegsOutput.ch1_delay_reset_o,
            Ch1ResValueLoad_i => WbRegsOutput.ch1_delay_reset_load_o,
            Ch2SetValue_ib => WbRegsOutput.ch2_delay_set_o,
            Ch2SetValueLoad_i => WbRegsOutput.ch2_delay_set_load_o,
            Ch2ResValue_ib => WbRegsOutput.ch2_delay_reset_o,
            Ch2ResValueLoad_i => WbRegsOutput.ch2_delay_reset_load_o,
            DelayValue_ob => DelayValue_ob,
            Ch1SetLe_o => Ch1SetLe_o,
            Ch1ResLe_o => Ch1ResLe_o,
            Ch2SetLe_o => Ch2SetLe_o,
            Ch2ResLe_o => Ch2ResLe_o,
            Busy_o => WbRegsInput.status_delay_configuration_busy_i
        );

    ----------------------------------
    -- output enable controls
    ----------------------------------
    Ch1OutputEnable_o <= WbRegsOutput.control_ch1_oe_o;
    Ch2OutputEnable_o <= WbRegsOutput.control_ch2_oe_o;
    WbRegsInput.status_channel_1_oe_i <= WbRegsOutput.control_ch1_oe_o;
    WbRegsInput.status_channel_2_oe_i <= WbRegsOutput.control_ch2_oe_o;

    ----------------------------------
    -- serial stream generators in CDC entity
    ----------------------------------
    cDelayedPulseGeneratorsCdc: entity work.DelayedPulseGeneratorsCdc(syn)
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Trigger_i => Trigger_i,
            Ch1SetMemAddress_ob11 => Ch1SetMemAddress_b11,
            Ch1SetMemData_ib32 => Ch1SetMemData_b32,
            Ch1SetMemReadStrobe_o => Ch1SetMemReadStrobe,
            Ch1ResMemAddress_ob11 => Ch1ResMemAddress_b11,
            Ch1ResMemData_ib32 => Ch1ResMemData_b32,
            Ch1ResMemReadStrobe_o => Ch1ResMemReadStrobe,
            Ch2SetMemAddress_ob11 => Ch2SetMemAddress_b11,
            Ch2SetMemData_ib32 => Ch2SetMemData_b32,
            Ch2SetMemReadStrobe_o => Ch2SetMemReadStrobe,
            Ch2ResMemAddress_ob11 => Ch2ResMemAddress_b11,
            Ch2ResMemData_ib32 => Ch2ResMemData_b32,
            Ch2ResMemReadStrobe_o => Ch2ResMemReadStrobe,
            Overflow_ib16 => WbRegsOutput.overflow_o,
            OverflowLoad_i => WbRegsOutput.overflow_load_o,
            TriggerLatency_ib16 => WbRegsOutput.trigger_latency_o,
            TriggerLatencyLoad_i => WbRegsOutput.trigger_latency_load_o,
            Ch1Mode_i => f_SlvToMode(WbRegsOutput.control_ch1_mode_o),
            Ch1ModeLoad_i => WbRegsOutput.control_ch1_mode_load_o,
            Ch1Running_o => WbRegsInput.status_channel_1_running_i,
            Ch2Mode_i => f_SlvToMode(WbRegsOutput.control_ch2_mode_o),
            Ch2ModeLoad_i => WbRegsOutput.control_ch2_mode_load_o,
            Ch2Running_o => WbRegsInput.status_channel_2_running_i,
            ClkRf_ik => ClkRf_k,
            Ch1SetStream_o => Ch1Set_o,
            Ch1ResetStream_o => Ch1Res_o,
            Ch2SetStream_o => Ch2Set_o,
            Ch2ResetStream_o => Ch2Res_o,
            Ch1FsmState_o => WbRegsInput.debug_i(2 downto 0),
            Ch2FsmState_o => WbRegsInput.debug_i(5 downto 3)
        );

    ----------------------------------
    -- LED outputs
    ----------------------------------
    
    cTriggerSyncer: entity work.Delay(syn)
        generic map (
            g_Width => 1,
            g_Delay => 3,
            g_AsyncRegUsed => "TRUE"
        )
        port map (
            Clk_ik => ClkRf_k,
            Data_ib(0) => Trigger_i,
            Data_ob(0) => TriggerRf
        );
    
    -- OUT1 - enabled and running
    LedSignal_b(1) <= WbRegsInput.status_channel_1_running_i and WbRegsOutput.control_ch1_oe_o;

    -- OUT2 - enabled and running
    LedSignal_b(2) <= WbRegsInput.status_channel_2_running_i and WbRegsOutput.control_ch2_oe_o;

    -- TRIG IN
    LedSignal_b(3) <= TriggerRf;

    -- CLK IN
    LedSignal_b(4) <= '1'; -- works just as "ON" LED as presence of ClkRf_ik cannot be sensed
    
    cSlowToggle: entity work.SlowToggle(behavioral)
        generic map (
            g_Width => 4,
            g_Ticks => g_ClkFrequency/15
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Test_i => WbRegsOutput.control_led_test_o,
            Signal_ib => LedSignal_b,
            Signal_ob => Led_ob
        );
    
    ----------------------------------
    -- Clock infrastructure
    ----------------------------------
    
    -- CLOCK_SELECTION [read/write]: Clock source selection
    --     0 (default): external clock used (connector on the front panel)
    --     1: FPGA loop clock used
    --     2: on-board VCXO clock used 
    
    Ad9512ClockSelection <=
        '0' when WbRegsOutput.control_clock_selection_o = "00" else   -- AD9512: CLK1
        '1';                                                          -- AD9512: CLK2
    Clk2Sel_o <=
        '0' when WbRegsOutput.control_clock_selection_o = "01" else   -- SY58017: FPGA LOOP
        '1';                                                          -- SY58017: VCXO

    cAd9512ClockSelectionChange: entity work.ChangeDetector(syn)
        port map (
            Clk_ik => Clk_ik,
            Signal_ib => WbRegsOutput.control_clock_selection_o,
            Change_o => Ad9512ClockSelectionChanged
        );

    Ad9512StartConfig <= WbRegsOutput.clock_ratio_m1_load_o or Ad9512ClockSelectionChanged;
    
    cAd9512Control: entity work.Ad9512Control(syn)
        generic map (
            g_ClkFrequency => g_ClkFrequency
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,

            Cfg_i => Ad9512StartConfig,
            Busy_o => WbRegsInput.status_clock_infrastructure_busy_i,
            
            ClockSelection_i => Ad9512ClockSelection,
            ClockRatioMinus1_ib => WbRegsOutput.clock_ratio_m1_o,

            SpiAd9512Sclk_o => SpiAd9512Sclk_o,
            SpiAd9512Mosi_o => SpiAd9512Mosi_o,
            SpiAd9512Miso_i => SpiAd9512Miso_i,
            SpiAd9512Cs_on => SpiAd9512Cs_on
        );
    
    cAd9512Syncer: entity work.Ad9512Syncer(syn)
        generic map (
            g_ClkFrequency => g_ClkFrequency
        )
        port map (
            Clk_ik => Clk_ik,
            StartSync_i => WbRegsOutput.control_ad9512_sync_o,
            Trigger_i => Trigger_i,
            Ad9512Sync_o => Ad9512SyncePulse
        );
    Ad9512Func_o <= not Ad9512SyncePulse;
    
    ClkOut_ok <= Clk_ik;
    
    ----------------------------------
    -- to be used
    ----------------------------------
    
    WbRegsInput.frequency_i <= (others => 'X');
    WbRegsInput.debug_i(31 downto 6) <= (others => '0');

end architecture;
