`include "timescale.svh"
`include "Config.svh"

package TestbenchPackage;

    import wishbone_b3_package::*;
    import Ad5600Package::*;
    import FfpgFmcPackage::*;
    import uvm_pkg::*;
    `include "uvm_macros.svh"

    typedef wishbone_b3_master_agent #(.ADR_W(`WB_ADDRESS_WIDTH), .DAT_W(`WB_DATA_WIDTH), .TAG_W(`WB_TAG_WIDTH)) td_wb_agent;
    typedef wishbone_b3_sequence_item #(.ADR_W(`WB_ADDRESS_WIDTH), .DAT_W(`WB_DATA_WIDTH), .TAG_W(`WB_TAG_WIDTH)) td_wb_tx;
        
    class UniversalSubscriber #(type t_Transaction) extends uvm_subscriber #(t_Transaction);
        `uvm_component_param_utils(UniversalSubscriber#(t_Transaction))

        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction

        function void write(t_Transaction t);
            `uvm_info("UNISUBS", $sformatf("Transaction \"%s\" processed: %s", t.get_type_name(), t.convert2string()), UVM_NONE);
        endfunction
        
    endclass
    
    class UniversalScoreboard #(type t_Transaction) extends uvm_component;
        `uvm_component_param_utils(UniversalScoreboard#(t_Transaction))

        uvm_analysis_export #(t_Transaction) AExport;
        uvm_analysis_export #(t_Transaction) BExport;

        uvm_tlm_analysis_fifo #(t_Transaction) AFifo;
        uvm_tlm_analysis_fifo #(t_Transaction) BFifo;
        
        int Count, Mismatches;

        function new(string name, uvm_component parent) ;
            super.new(name, parent);
            Count = 0;
            Mismatches = 0;
        endfunction

        function void build_phase(uvm_phase phase);
            AFifo = new("AFifo", this);
            BFifo = new("BFifo", this);
            AExport = new("AExport", this);
            BExport = new("BExport", this);
        endfunction

        function void connect_phase(uvm_phase phase);
            AExport.connect(AFifo.analysis_export);
            BExport.connect(BFifo.analysis_export);
        endfunction

        task run_phase(uvm_phase phase);
            string s;
            t_Transaction ATx;
            t_Transaction BTx;
            forever begin
                BFifo.peek(BTx);
                AFifo.get(ATx);
                BFifo.get(BTx);
                if (!ATx.compare(BTx)) begin
                    $sformat(s, "Two transactions \"%s\" are not the same!\n    A: %s\n    B: %s", t_Transaction::type_name, ATx.convert2string(), BTx.convert2string());
                    `uvm_error("UniversalScoreboard", s)
                    Mismatches++;
                end
                Count++;
            end
        endtask

        function void report_phase(uvm_phase phase);
            `uvm_info("UniversalScoreboard", $sformatf("%s: Count:      %0d", t_Transaction::type_name, Count), UVM_LOW)
            if (Count > 0)
                `uvm_info("UniversalScoreboard", $sformatf("%s: Mismatches: %0d (%f %%)", t_Transaction::type_name, Mismatches, 100*(Mismatches/$itor(Count))), UVM_LOW)
            if (~AFifo.is_empty())
                `uvm_error("UniversalScoreboard", $sformatf("There are some transactions \"%s\" left from A side! (count: %0d)", t_Transaction::type_name, AFifo.used()))
            if (~BFifo.is_empty())
                `uvm_error("UniversalScoreboard", $sformatf("There are some transactions \"%s\" left from B side! (count: %0d)", t_Transaction::type_name, BFifo.used()))
        endfunction

    endclass

    class PulseScoreboard extends uvm_component;
        `uvm_component_utils(PulseScoreboard)

        uvm_analysis_export #(FfpgFmcPulseTransaction) SetPulseExport;
        uvm_analysis_export #(FfpgFmcPulseTransaction) MonitoredPulseExport;

        uvm_tlm_analysis_fifo #(FfpgFmcPulseTransaction) SetPulseFifo;
        uvm_tlm_analysis_fifo #(FfpgFmcPulseTransaction) MonitoredPulseFifo;
        
        FfpgFmcPulseTransaction LastPulseConfiguration = null;
        
        int Count, Mismatches;

        function new(string name, uvm_component parent) ;
            super.new(name, parent);
            Count = 0;
            Mismatches = 0;
        endfunction

        function void build_phase(uvm_phase phase);
            SetPulseFifo = new("SetPulseFifo", this);
            MonitoredPulseFifo = new("MonitoredPulseFifo", this);
            SetPulseExport = new("SetPulseExport", this);
            MonitoredPulseExport = new("MonitoredPulseExport", this);
        endfunction

        function void connect_phase(uvm_phase phase);
            SetPulseExport.connect(SetPulseFifo.analysis_export);
            MonitoredPulseExport.connect(MonitoredPulseFifo.analysis_export);
        endfunction
        
        protected task ReceivePulseConfiguration;
            FfpgFmcPulseTransaction Tx;
            
            forever begin
                SetPulseFifo.get(Tx);
                this.LastPulseConfiguration = Tx;
            end
        endtask
        
        protected task ReceiveMonitoredPulse;
            FfpgFmcPulseTransaction Tx;
            string s;
            
            forever begin
                MonitoredPulseFifo.get(Tx);
                Count++;
                if (this.LastPulseConfiguration == null) begin
                    `uvm_error("PulseScoreboard", "Received a pulse from monitor, but no configuration has been received! (mismatch)")
                    Mismatches++;
                end else if (!Tx.compare(this.LastPulseConfiguration)) begin
                    $sformat(s,
                        "A pulse received from the monitor doesn't conform to the last configuration!\n    Configuration:  %s\n    Received pulse: %s",
                        this.LastPulseConfiguration.convert2string(), Tx.convert2string());
                    `uvm_error("PulseScoreboard", s)
                    Mismatches++;
                end
            end
        endtask

        task run_phase(uvm_phase phase);
            fork
                this.ReceivePulseConfiguration();
                this.ReceiveMonitoredPulse();
            join
        endtask

        function void report_phase(uvm_phase phase);
            `uvm_info("PulseScoreboard", $sformatf("Count:      %0d", Count), UVM_LOW)
            if (Count > 0)
                `uvm_info("PulseScoreboard", $sformatf("Mismatches: %0d (%f %%)", Mismatches, 100*(Mismatches/$itor(Count))), UVM_LOW)
        endfunction

    endclass

    class DacScoreboard #(int WbAddress) extends uvm_component;
        // adopted from https://verificationacademy.com/cookbook/scoreboards
        `uvm_component_param_utils(DacScoreboard#(WbAddress))

        uvm_analysis_export #(td_wb_tx) WbExport;
        uvm_analysis_export #(Ad5600Transaction) AdExport;

        uvm_tlm_analysis_fifo #(td_wb_tx) WbFifo;
        uvm_tlm_analysis_fifo #(Ad5600Transaction) AdFifo;
        
        int Count, Mismatches;

        function new(string name, uvm_component parent) ;
            super.new(name, parent);
            Count = 0;
            Mismatches = 0;
        endfunction

        function void build_phase(uvm_phase phase);
            WbFifo = new("WbFifo", this);
            AdFifo = new("AdFifo", this);
            WbExport = new("WbExport", this);
            AdExport = new("AdExport", this);
        endfunction

        function void connect_phase(uvm_phase phase);
            WbExport.connect(WbFifo.analysis_export);
            AdExport.connect(AdFifo.analysis_export);
        endfunction

        task run_phase(uvm_phase phase);
            string s;
            td_wb_tx WbTx;
            Ad5600Transaction AdTx;
            forever begin
                forever begin
                    // wait for specific command
                    WbFifo.get(WbTx);
                    if (WbTx.direction_e == WB_B3_DIR_WRITE & WbTx.address == WbAddress & WbTx.response_e == WB_B3_RESPONSE_ACK_OK)
                        break;
                end
                AdFifo.get(AdTx);
                if (WbTx.data != AdTx.VoltageRegister) begin
                    $sformat(s, "Data written to DAC \"%s\" (0x%H) does not match data received on DAC (0x%H)", AdTx.Chip.Name, WbTx.data, AdTx.VoltageRegister);
                    `uvm_error("DacScoreboard", s)
                    Mismatches++;
                end
                Count++;
            end
        endtask

        function void report_phase(uvm_phase phase);
            `uvm_info("DacScoreboard", $sformatf("Count:      %0d", Count), UVM_LOW)
            if (Count > 0)
                `uvm_info("DacScoreboard", $sformatf("Mismatches: %0d (%f %%)", Mismatches, 100*(Mismatches/$itor(Count))), UVM_LOW)
            if (~AdFifo.is_empty())
                `uvm_error("DacScoreboard", $sformatf("There are some transactions left from DAC! (count: %0d)", AdFifo.used()))
        endfunction

    endclass

    class TestEnvorinment extends uvm_env;
        `uvm_component_utils(TestEnvorinment)
        
        rand int ClockPeriod; // in time units
        rand int TriggerPeriodicity; // in clock cycles
        rand int TriggerPhase; // in time units
        rand int TriggerLatency; // in clock cycles

        constraint ClockPeriodConstrain {
            ClockPeriod inside {[2:200]};
            ClockPeriod % 2 == 0;
        }

        constraint TriggerPeriodicityConstrain {
            TriggerPeriodicity inside {[10:1000]};
        }

        constraint TriggerPhaseConstrain {
            TriggerPhase inside {[0:ClockPeriod-1]};
        }
        
        constraint TriggerLatencyConstraint {
            TriggerLatency inside {[0:20]};
        }
        
        Ad5600Chip TriggerDacChip = new(AD5600_TYPE_5660, AD5600_SUFFIX_3, 5, "Trigger threshold");
        Ad5600Chip VcxoDacChip = new(AD5600_TYPE_5660, AD5600_SUFFIX_3, 5, "VCXO");

        UVM_FILE file_h;
        
        virtual FfpgInterface LocalInterface;

        td_wb_agent WbAgent;
        //wishbone_b3_reg_adapter #(.ADR_W(`WB_ADDRESS_WIDTH), .DAT_W(`WB_DATA_WIDTH), .TAG_W(`WB_TAG_WIDTH)) wb_b3_reg_adapter;
        
        Ad5600Agent TriggerDacAgent;
        Ad5600Agent VcxoDacAgent;
        
        FfpgFmcAgent FmcAgent;
        
        UniversalSubscriber #(td_wb_tx) WbSubscriber;
        UniversalSubscriber #(Ad5600Transaction) TriggerDacSubscriber;
        UniversalSubscriber #(Ad5600Transaction) VcxoDacSubscriber;
        UniversalSubscriber #(FfpgFmcPulseTransaction) FmcPulseSubscriber;
        
        DacScoreboard #(.WbAddress(`GET_WB_FFPG_ADDR(`ADDR_FFPG_TRIGGER_THRESHOLD))) TriggerDacScoreboard;
        DacScoreboard #(.WbAddress(`GET_WB_FFPG_ADDR(`ADDR_FFPG_VCXO_VOLTAGE))) VcxoDacScoreboard;
        
        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction

        function void build_phase(uvm_phase phase);
            wishbone_b3_master_cfg WbConfig = new();
            Ad5600Config TriggerDacConfig = new();
            Ad5600Config VcxoDacConfig = new();
            FfpgFmcConfig FmcConfig = new();
            
            if(!uvm_config_db #(virtual FfpgInterface)::get(this, "", "dut_vi", LocalInterface))
                `uvm_fatal("CONF", "No DUT virtual interface set")

            WbConfig.is_active = UVM_ACTIVE;
            WbAgent = td_wb_agent::type_id::create("WbAgent", this);
            WbAgent.cfg  = WbConfig;
            WbAgent.sigs = LocalInterface.Wb;
            
            TriggerDacConfig.VirtualInterface = LocalInterface.TriggerDac;
            TriggerDacConfig.Chip = TriggerDacChip;
            TriggerDacAgent = Ad5600Agent::type_id::create("TriggerDacAgent", this);
            uvm_config_db #(Ad5600Config)::set(TriggerDacAgent, "*", "Ad5600Config", TriggerDacConfig);
            
            VcxoDacConfig.VirtualInterface = LocalInterface.VcxoDac;
            VcxoDacConfig.Chip = VcxoDacChip;
            VcxoDacAgent = Ad5600Agent::type_id::create("VcxoDacAgent", this);
            uvm_config_db #(Ad5600Config)::set(VcxoDacAgent, "*", "Ad5600Config", VcxoDacConfig);
            
            FmcConfig.VirtualInterface = LocalInterface.Fmc;
            FmcConfig.ClockPeriod = this.ClockPeriod;
            FmcConfig.TriggerPeriodicity = this.TriggerPeriodicity;
            FmcConfig.TriggerPhase = this.TriggerPhase;
            FmcConfig.TriggerLatency = this.TriggerLatency;
            FmcAgent = FfpgFmcAgent::type_id::create("FmcAgent", this);
            uvm_config_db #(FfpgFmcConfig)::set(FmcAgent, "", "FfpgFmcConfig", FmcConfig); // for Agent
            uvm_config_db #(FfpgFmcConfig)::set(FmcAgent, "*", "FfpgFmcConfig", FmcConfig); // for all Agent's children
            
            WbSubscriber = UniversalSubscriber #(td_wb_tx)::type_id::create("WbSubscriber", this);
            TriggerDacSubscriber = UniversalSubscriber #(Ad5600Transaction)::type_id::create("TriggerDacSubscriber", this);
            VcxoDacSubscriber = UniversalSubscriber #(Ad5600Transaction)::type_id::create("VcxoDacSubscriber", this);
            FmcPulseSubscriber = UniversalSubscriber #(FfpgFmcPulseTransaction)::type_id::create("FmcPulseSubscriber", this);
            
            TriggerDacScoreboard = DacScoreboard #(.WbAddress(`GET_WB_FFPG_ADDR(`ADDR_FFPG_TRIGGER_THRESHOLD)))::type_id::create("TriggerDacScoreboard", this);
            VcxoDacScoreboard = DacScoreboard #(.WbAddress(`GET_WB_FFPG_ADDR(`ADDR_FFPG_VCXO_VOLTAGE)))::type_id::create("VcxoDacScoreboard", this);

            file_h = $fopen("uvm.log", "a");
            uvm_top.set_report_default_file_hier(file_h);
            uvm_top.set_report_severity_action_hier(UVM_INFO, UVM_DISPLAY | UVM_LOG);
        endfunction
        
        function void connect_phase(uvm_phase phase);
            WbAgent.analysis_port.connect(WbSubscriber.analysis_export);
            // // connect RAL accessibility to this port
            // wb_b3_reg_adapter = wishbone_b3_master_agent #(.ADR_W(`WB_ADDRESS_WIDTH), .DAT_W(`WB_DATA_WIDTH), .TAG_W(`WB_TAG_WIDTH))::type_id::create("wb_b3_reg_adapter");
            //    // above line should be new() of wishbone_b3_reg_adapter?
            // regmodel.default_map.set_sequencer(WbAgent.seqr, wb_b3_reg_adapter);
            TriggerDacAgent.AnalysisPort.connect(TriggerDacSubscriber.analysis_export);
            VcxoDacAgent.AnalysisPort.connect(VcxoDacSubscriber.analysis_export);
            FmcAgent.AnalysisPort.connect(FmcPulseSubscriber.analysis_export);
            
            WbAgent.analysis_port.connect(TriggerDacScoreboard.WbExport);
            TriggerDacAgent.AnalysisPort.connect(TriggerDacScoreboard.AdExport);

            WbAgent.analysis_port.connect(VcxoDacScoreboard.WbExport);
            VcxoDacAgent.AnalysisPort.connect(VcxoDacScoreboard.AdExport);
        endfunction

        function void start_of_simulation_phase(uvm_phase phase);
        endfunction

    endclass
    
    class ReportingSequence #(type t_Transaction = uvm_sequence_item) extends uvm_sequence #(t_Transaction);
        
        function new (string name);
            super.new(name);
        endfunction

        virtual task pre_body();  
            `uvm_info("REPSEQ", $sformatf("Sequence \"%s\" is being processed: %s", this.get_type_name(), this.convert2string()), UVM_NONE);
        endtask

    endclass

    class SeqSetDac extends ReportingSequence #(td_wb_tx);
        `uvm_object_utils(SeqSetDac)
        
        rand int VoltageRegister;
        rand int RepeatCount;
        Ad5600Chip Chip;
        int WbDacAddress;
        int WbStatusBit;

        constraint VoltageRegisterConstrain {
            VoltageRegister >= 0;
            VoltageRegister < 2**Chip.getDataBits();
        }
        
        constraint RepeatCountConstraint {
            RepeatCount inside {[5:20]};
        }

        function new (string name = "");
            super.new(name);
        endfunction

        task body;
            td_wb_tx tx;
            bit busy = 0;
            
            repeat (RepeatCount) begin
                busy = 1;
                repeat(3) begin
                    tx = td_wb_tx::type_id::create("tx");
                    start_item(tx);
                    assert(tx.randomize());
                    tx.direction_e = WB_B3_DIR_READ;
                    tx.address = `GET_WB_FFPG_ADDR(`ADDR_FFPG_STATUS);
                    finish_item(tx);
                    if (tx.response_e == WB_B3_RESPONSE_ACK_OK && tx.data[WbStatusBit] == 0) begin
                        busy = 0;
                        break;
                    end
                    #500;
                end

                if (~busy) begin
                    tx = td_wb_tx::type_id::create("tx");
                    start_item(tx);
                    assert(tx.randomize());
                    tx.direction_e = WB_B3_DIR_WRITE;
                    tx.address = WbDacAddress;
                    tx.data = VoltageRegister;
                    assert(randomize(VoltageRegister));
                    finish_item(tx);
                end else
                    `uvm_warning("TB", "DAC is busy for too long!")
                
                #100;
            end

        endtask

    endclass

    class SeqSetDelay extends ReportingSequence #(td_wb_tx);
        `uvm_object_utils(SeqSetDelay)
        
        typedef enum {CH1_SET, CH1_RESET, CH2_SET, CH2_RESET} t_Output;
        
        rand t_Output ActiveOutput;
        rand logic [9:0] Delay;
        rand int RepeatCount;

        constraint RepeatCountConstraint {
            RepeatCount inside {[5:20]};
        }

        function new (string name = "");
            super.new(name);
        endfunction

        function string convert2string;
            return $sformatf("Setting of delay for output %0d: delay = %0d (0x%H), repeat count = %0d.", this.ActiveOutput, this.Delay, this.Delay, this.RepeatCount);
        endfunction

        task body;
            td_wb_tx tx;
            bit busy = 0;
            int WbDelayAddress;
            
            case(ActiveOutput)
                CH1_SET: WbDelayAddress = `GET_WB_FFPG_ADDR(`ADDR_FFPG_CH1_DELAY_SET);
                CH1_RESET: WbDelayAddress = `GET_WB_FFPG_ADDR(`ADDR_FFPG_CH1_DELAY_RESET);
                CH2_SET: WbDelayAddress = `GET_WB_FFPG_ADDR(`ADDR_FFPG_CH2_DELAY_SET);
                CH2_RESET: WbDelayAddress = `GET_WB_FFPG_ADDR(`ADDR_FFPG_CH2_DELAY_RESET);
            endcase

            repeat (RepeatCount) begin
                busy = 1;
                repeat(2) begin
                    tx = td_wb_tx::type_id::create("tx");
                    start_item(tx);
                    assert(tx.randomize());
                    tx.direction_e = WB_B3_DIR_READ;
                    tx.address = `GET_WB_FFPG_ADDR(`ADDR_FFPG_STATUS);
                    finish_item(tx);
                    if (tx.response_e == WB_B3_RESPONSE_ACK_OK && tx.data[`FFPG_STATUS_DELAY_CONFIGURATION_BUSY_OFFSET] == 0) begin
                        busy = 0;
                        break;
                    end
                    #10;
                end

                if (~busy) begin
                    tx = td_wb_tx::type_id::create("tx");
                    start_item(tx);
                    assert(tx.randomize());
                    tx.direction_e = WB_B3_DIR_WRITE;
                    tx.address = WbDelayAddress;
                    tx.data = Delay;
                    assert(randomize(Delay));
                    assert(randomize(ActiveOutput));
                    finish_item(tx);
                end else
                    `uvm_warning("TB", "Delay configuration is busy for too long!")
                
                #100;
            end

        endtask

    endclass

    class SeqSetChannel extends ReportingSequence #(td_wb_tx);
        `uvm_object_utils(SeqSetChannel)
        
        int TriggerLatency;
        int TriggerPeriodicity;
        rand int Channel;
        rand int Overflow;
        rand logic [`WB_DATA_WIDTH-1:0] SetData [0:(`STREAM_SIZE/`WB_DATA_WIDTH)-1];
        rand logic [`WB_DATA_WIDTH-1:0] ResData [0:(`STREAM_SIZE/`WB_DATA_WIDTH)-1];
        
        constraint ChannelConstraint {
            Channel inside {[1:2]};
        }
        
        constraint OverflowConstraint {
            Overflow inside {[0:`STREAM_SIZE-1]};
            Overflow > TriggerLatency;
            Overflow <= TriggerPeriodicity;
        }
        
        function new (string name = "");
            super.new(name);
        endfunction

        function string convert2string;
            return $sformatf("Setting for channel %0d: overflow = %0d (0x%H), trigger latency = %0d (0x%H).", this.Channel, this.Overflow, this.Overflow, this.TriggerLatency, this.TriggerLatency);
        endfunction
        
        function int getOverflowInWords;
            return (Overflow / `WB_DATA_WIDTH) + 1;
        endfunction

        task body;
            td_wb_tx tx;
            int SetMemBaseAddress = (Channel == 1) ? `GET_WB_FFPG_ADDR(`BASE_FFPG_CH1_SET_MEM) : `GET_WB_FFPG_ADDR(`BASE_FFPG_CH2_SET_MEM);
            int ResMemBaseAddress = (Channel == 1) ? `GET_WB_FFPG_ADDR(`BASE_FFPG_CH1_RES_MEM) : `GET_WB_FFPG_ADDR(`BASE_FFPG_CH2_RES_MEM);
            int OverflowInWords = getOverflowInWords();
            
            // configure overflow
            tx = td_wb_tx::type_id::create("tx");
            start_item(tx);
            assert(tx.randomize());
            tx.direction_e = WB_B3_DIR_WRITE;
            tx.address = `GET_WB_FFPG_ADDR(`ADDR_FFPG_OVERFLOW);
            tx.data = Overflow;
            finish_item(tx);
            assert(tx.response_e == WB_B3_RESPONSE_ACK_OK);
            
            // configure trigger latency
            tx = td_wb_tx::type_id::create("tx");
            start_item(tx);
            assert(tx.randomize());
            tx.direction_e = WB_B3_DIR_WRITE;
            tx.address = `GET_WB_FFPG_ADDR(`ADDR_FFPG_TRIGGER_LATENCY);
            tx.data = this.TriggerLatency;
            finish_item(tx);
            assert(tx.response_e == WB_B3_RESPONSE_ACK_OK);
            
            // configure Set and Res data
            for (int i = 0; i < OverflowInWords; i++) begin
                // SetData
                tx = td_wb_tx::type_id::create("tx");
                start_item(tx);
                assert(tx.randomize());
                tx.direction_e = WB_B3_DIR_WRITE;
                tx.address = SetMemBaseAddress + i*4;
                tx.data = SetData[i];
                finish_item(tx);
                assert(tx.response_e == WB_B3_RESPONSE_ACK_OK);
                // ResData
                tx = td_wb_tx::type_id::create("tx");
                start_item(tx);
                assert(tx.randomize());
                tx.direction_e = WB_B3_DIR_WRITE;
                tx.address = ResMemBaseAddress + i*4;
                tx.data = ResData[i];
                finish_item(tx);
                assert(tx.response_e == WB_B3_RESPONSE_ACK_OK);
            end
            
            // #100; // ??

        endtask

    endclass
    
    class SeqStartChannel extends ReportingSequence #(td_wb_tx);
        `uvm_object_utils(SeqStartChannel)
        
        rand int Channel;
        
        constraint ChannelConstraint {
            Channel inside {[1:2]};
        }
        
        function new (string name = "");
            super.new(name);
        endfunction

        function string convert2string;
            return $sformatf("Starting channel %0d.", this.Channel);
        endfunction

        task body;
            td_wb_tx tx;
            int Position = (Channel == 1) ? `FFPG_CONTROL_CH1_MODE_OFFSET : `FFPG_CONTROL_CH2_MODE_OFFSET;
            
            tx = td_wb_tx::type_id::create("tx");
            start_item(tx);
            assert(tx.randomize());
            tx.direction_e = WB_B3_DIR_WRITE;
            tx.address = `GET_WB_FFPG_ADDR(`ADDR_FFPG_CONTROL);
            tx.data = 1 << Position;
            finish_item(tx);
            assert(tx.response_e == WB_B3_RESPONSE_ACK_OK);
        endtask

    endclass
    
    class SeqGenerateBeamSignals extends ReportingSequence #(FfpgFmcDriverTransaction);
        `uvm_object_utils(SeqGenerateBeamSignals)
        
        int Overflow;
        int Cycles = -1; // overides random Cycles the transaction iff > -1
        
        function new (string name = "");
            super.new(name);
        endfunction

        function string convert2string;
            return "Beam signal is being generated.";
        endfunction

        task body;
            FfpgFmcDriverTransaction tx;
            
            tx = FfpgFmcDriverTransaction::type_id::create("tx");
            start_item(tx);
            if (this.Cycles > -1)
                assert(tx.randomize() with {Cycles == local::Cycles;});
            else
                assert(tx.randomize());
            finish_item(tx);
        endtask

    endclass

    class SeqGeneratePulses extends ReportingSequence;
        `uvm_object_utils(SeqGeneratePulses)
        
        uvm_sequencer #(td_wb_tx) WbSequencer;
        uvm_sequencer #(FfpgFmcDriverTransaction) FmcSequencer;
        uvm_analysis_port #(FfpgFmcPulseTransaction) GenerationPort;
        int TriggerLatency;
        int TriggerPeriodicity;
        rand int RepeatCount;
        
        constraint CountConstraint {
            RepeatCount inside {[1:20]};
        }
        
        function new (string name = "");
            super.new(name);
        endfunction

        function string convert2string;
            return $sformatf("Generating %0d pulses.", this.RepeatCount);
        endfunction

        task body;
            FfpgFmcPulseTransaction PulseTx;
            SeqSetChannel SeqSet;
            SeqStartChannel SeqStart;
            SeqGenerateBeamSignals SeqGenerate;
            int Channel;
            int OverflowInWords;
            int SetDataOffset, ResDataOffset;
            int SetDataMask, ResDataMask;            
            
            repeat (RepeatCount) begin
                PulseTx = FfpgFmcPulseTransaction::type_id::create("PulseTx");
                PulseTx.BunchCount = this.TriggerPeriodicity;
                assert(PulseTx.randomize()
                    with { // XXX
                        Channel == 2;
                        StartBunch == 978;
                        StopBunch == 978;
                    }
                );
                Channel = PulseTx.Channel;
                GenerationPort.write(PulseTx);
                
                // configure channel
                SeqSet = SeqSetChannel::type_id::create("SeqSet");
                SeqSet.TriggerLatency = this.TriggerLatency;
                SeqSet.TriggerPeriodicity = this.TriggerPeriodicity;
                assert(SeqSet.randomize() with {Overflow == TriggerPeriodicity;});
                SeqSet.Channel = Channel;
                OverflowInWords = SeqSet.getOverflowInWords();
                // clear Set and Res data
                for (int i = 0; i < OverflowInWords; i++) begin
                    SeqSet.SetData[i] = 0;
                    SeqSet.ResData[i] = 0;
                end
                // configure Set and Res data
                SetDataOffset = PulseTx.StartBunch / 32;
                ResDataOffset = PulseTx.StopBunch / 32;
                SetDataMask = 1 << (PulseTx.StartBunch % 32);
                ResDataMask = 1 << (PulseTx.StopBunch % 32);
                SeqSet.SetData[SetDataOffset] = SetDataMask;
                SeqSet.ResData[ResDataOffset] = ResDataMask;
                SeqSet.start(WbSequencer, this);

                // start channel
                SeqStart = SeqStartChannel::type_id::create("SeqStart");
                assert(SeqStart.randomize());
                SeqStart.Channel = Channel;
                SeqStart.start(WbSequencer, this);

                // generate FMC input signals
                SeqGenerate = SeqGenerateBeamSignals::type_id::create("SeqGenerate");
                SeqGenerate.Overflow = SeqSet.Overflow;
                SeqGenerate.Cycles = 5;
                SeqGenerate.start(FmcSequencer, this);

            end
        endtask
        
    endclass
    
    class TestBase extends uvm_test;
        `uvm_component_utils(TestBase)

        TestEnvorinment Env_h;   

        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction: new

        function void build_phase(uvm_phase phase);
            Env_h = TestEnvorinment::type_id::create("Env_h", this);
            assert(Env_h.randomize()
                with { // XXX
                    ClockPeriod == 2;
                    TriggerPeriodicity == 990;
                    TriggerPhase == 0;
                    TriggerLatency == 8;
                }
            );
        endfunction

    endclass
    
    class TestTriggerAdc extends TestBase;
        `uvm_component_utils(TestTriggerAdc)

        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction

        function void build_phase(uvm_phase phase);
            super.build_phase(phase);
        endfunction

        task run_phase(uvm_phase phase);
            SeqSetDac seq;
            phase.raise_objection(this);
            #100;
            
            // test trigger DAC
            seq = SeqSetDac::type_id::create("seq");
            seq.Chip = Env_h.TriggerDacChip;
            seq.WbDacAddress = `GET_WB_FFPG_ADDR(`ADDR_FFPG_TRIGGER_THRESHOLD);
            seq.WbStatusBit = `FFPG_STATUS_DAC_TRIGGER_BUSY_OFFSET;
            assert(seq.randomize());
            seq.start(Env_h.WbAgent.seqr);
            
            // test VCXO DAC
            seq = SeqSetDac::type_id::create("seq");
            seq.Chip = Env_h.VcxoDacChip;
            seq.WbDacAddress = `GET_WB_FFPG_ADDR(`ADDR_FFPG_VCXO_VOLTAGE);
            seq.WbStatusBit = `FFPG_STATUS_DAC_VCXO_BUSY_OFFSET;
            assert(seq.randomize());
            seq.start(Env_h.WbAgent.seqr);
            
            phase.drop_objection(this);
        endtask

    endclass
    
    class TestDelayConfiguration extends TestBase;
        `uvm_component_utils(TestDelayConfiguration)

        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction

        function void build_phase(uvm_phase phase);
            super.build_phase(phase);
        endfunction

        task run_phase(uvm_phase phase);
            SeqSetDelay seq;
            phase.raise_objection(this);
            #100;
            
            seq = SeqSetDelay::type_id::create("seq");
            assert(seq.randomize());
            seq.start(Env_h.WbAgent.seqr);
            
            phase.drop_objection(this);
        endtask

    endclass

    class TestPulseGenerationBlind extends TestBase;
        `uvm_component_utils(TestPulseGenerationBlind)

        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction

        function void build_phase(uvm_phase phase);
            super.build_phase(phase);
        endfunction

        task run_phase(uvm_phase phase);
            SeqSetChannel SeqSet;
            SeqStartChannel SeqStart;
            SeqGenerateBeamSignals SeqGenerate;
            
            phase.raise_objection(this);
            #100;
            
            // configure channel
            SeqSet = SeqSetChannel::type_id::create("SeqSet");
            SeqSet.TriggerLatency = Env_h.TriggerLatency;
            SeqSet.TriggerPeriodicity = Env_h.TriggerPeriodicity;
            assert(SeqSet.randomize());
            SeqSet.start(Env_h.WbAgent.seqr);

            #100;
            
            // start channel
            SeqStart = SeqStartChannel::type_id::create("SeqStart");
            assert(SeqStart.randomize());
            SeqStart.Channel = SeqSet.Channel;
            SeqStart.start(Env_h.WbAgent.seqr);
            
            //#1000;
            SeqGenerate = SeqGenerateBeamSignals::type_id::create("SeqGenerate");
            SeqGenerate.Overflow = SeqSet.Overflow;
            SeqGenerate.start(Env_h.FmcAgent.Sequencer_h);
            
            phase.drop_objection(this);
        endtask

    endclass

    class TestPulseGenerationSequential extends TestBase;
        `uvm_component_utils(TestPulseGenerationSequential)

        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction

        function void build_phase(uvm_phase phase);
            super.build_phase(phase);
        endfunction

        task run_phase(uvm_phase phase);
            SeqSetChannel SeqSet;
            SeqStartChannel SeqStart;
            SeqGenerateBeamSignals SeqGenerate;
            int OverflowInWords;
            int TestChannel = 1;
            
            phase.raise_objection(this);
            #100;
            
            // configure channel
            SeqSet = SeqSetChannel::type_id::create("SeqSet");
            SeqSet.TriggerLatency = Env_h.TriggerLatency;
            SeqSet.TriggerPeriodicity = Env_h.TriggerPeriodicity;
            assert(SeqSet.randomize() with {Overflow == TriggerPeriodicity;});
            SeqSet.Channel = TestChannel;
            OverflowInWords = SeqSet.getOverflowInWords();
            // configure Set and Res data
            for (int i = 0; i < OverflowInWords; i++) begin
                SeqSet.SetData[i] = 'h5555_5555;
                SeqSet.ResData[i] = (i%2) * -1;
            end
            SeqSet.start(Env_h.WbAgent.seqr);

            #100;
            
            // start channel
            SeqStart = SeqStartChannel::type_id::create("SeqStart");
            assert(SeqStart.randomize());
            SeqStart.Channel = TestChannel;
            SeqStart.start(Env_h.WbAgent.seqr);
            
            //#1000;
            SeqGenerate = SeqGenerateBeamSignals::type_id::create("SeqGenerate");
            SeqGenerate.Overflow = SeqSet.Overflow;
            SeqGenerate.start(Env_h.FmcAgent.Sequencer_h);
            
            phase.drop_objection(this);
        endtask

    endclass

    class TestPulseGeneration extends TestBase;
        `uvm_component_utils(TestPulseGeneration)

        UniversalSubscriber #(FfpgFmcPulseTransaction) PulseSubscriber;
        PulseScoreboard PulseScoreboard_h;
        uvm_analysis_port #(FfpgFmcPulseTransaction) GenerationPort;
        
        function new(string name, uvm_component parent);
            super.new(name, parent);
        endfunction

        function void build_phase(uvm_phase phase);
            super.build_phase(phase);
            PulseSubscriber = UniversalSubscriber #(FfpgFmcPulseTransaction)::type_id::create("PulseSubscriber", this);
            PulseScoreboard_h = PulseScoreboard::type_id::create("PulseScoreboard_h", this);
            GenerationPort = new("GenerationPort", this);
        endfunction

        function void connect_phase(uvm_phase phase);
            GenerationPort.connect(PulseScoreboard_h.SetPulseExport);
            GenerationPort.connect(PulseSubscriber.analysis_export);
            Env_h.FmcAgent.AnalysisPort.connect(PulseScoreboard_h.MonitoredPulseExport);
        endfunction

        task run_phase(uvm_phase phase);
            SeqGeneratePulses SeqGen;
            
            phase.raise_objection(this);
            #100;
            
            SeqGen = SeqGeneratePulses::type_id::create("SeqGen");
            SeqGen.TriggerLatency = Env_h.TriggerLatency;
            SeqGen.TriggerPeriodicity = Env_h.TriggerPeriodicity;
            assert(SeqGen.randomize());
            SeqGen.RepeatCount = 1; // XXX
            SeqGen.WbSequencer = Env_h.WbAgent.seqr;
            SeqGen.FmcSequencer = Env_h.FmcAgent.Sequencer_h;
            SeqGen.GenerationPort = this.GenerationPort;
            SeqGen.start(null);

            phase.drop_objection(this);
        endtask

    endclass

endpackage