library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity HeartBeat is
    generic (
        g_ClkFrequency: positive -- input clock frequency in Hz
    );
    port (
        Clk_ik: in std_logic;
        Reset_ir: in std_logic;
        HeartBeat_o: out std_logic := '0'
    );
end entity;


architecture syn of HeartBeat is

    --! computes least possible width of the vector to store value X
    -- http://stackoverflow.com/a/12751341/615627
    function f_log2(x: natural) return natural is
        variable i: natural;
    begin
        i := 0;
        while (2**i < x) loop
            i := i + 1;
        end loop;
        return i;
    end function;
    
    function f_min(a, b: integer) return integer is
    begin
        if a < b then
            return a;
        else
            return b;
        end if;
    end function;
    
    function f_max(a, b: integer) return integer is
    begin
        if a > b then
            return a;
        else
            return b;
        end if;
    end function;

    constant c_ClkFreqWidth: positive := f_log2(g_ClkFrequency);
    constant c_PwmWidth: positive := f_max(f_min(c_ClkFreqWidth, 12), 4);
    constant c_DividerWidth: natural := c_ClkFreqWidth - c_PwmWidth;
    constant c_DividerLimit: natural := g_ClkFrequency / (2**c_PwmWidth);

    signal PwmCountDown: std_logic := '0';
    signal UpdatePwmValue, PwmCounterOverflow: std_logic;
    signal PwmCounter_b, PwmValue_b, PwmValue_bd: signed(c_PwmWidth-1 downto 0) := (others => '0');
    
begin

    gNeedDivider: if c_DividerWidth > 0 generate
        cDivider: entity work.Counter(syn)
            generic map (
                g_Width => c_DividerWidth,
                g_Limit => c_DividerLimit
            )
            port map (
                Clk_ik => Clk_ik,
                Reset_ir => Reset_ir,
                Enable_i => '1',
                Set_i => '0',
                SetValue_ib => (others => '0'),
                Overflow_o => UpdatePwmValue,
                Value_ob => open
            );
    end generate;
    
    gDoesntNeedDivider: if c_DividerWidth = 0 generate
        UpdatePwmValue <= '1';
    end generate;
    
    pPwmValue: process (Clk_ik) begin
        if rising_edge(Clk_ik) then
            if Reset_ir = '1' then
                PwmValue_b <= (others => '0');
                PwmCountDown <= '0';
            elsif UpdatePwmValue = '1' then
                if PwmCountDown = '1' then
                    if PwmValue_b = to_signed(1, c_PwmWidth) then
                        PwmCountDown <= '0';
                        PwmValue_b <= PwmValue_b + 1;
                    else
                        PwmValue_b <= PwmValue_b - 1;
                    end if;
                else
                    if PwmValue_b = to_signed(-1, c_PwmWidth) then
                        PwmCountDown <= '1';
                        PwmValue_b <= PwmValue_b - 1;
                    else
                        PwmValue_b <= PwmValue_b + 1;
                    end if;
                end if;
            end if;
        end if;
    end process;
    
    cPwmCounter: entity work.Counter(syn)
        generic map (
            g_Width => c_PwmWidth
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => Reset_ir,
            Enable_i => '1',
            Set_i => '0',
            SetValue_ib => (others => '0'),
            Overflow_o => PwmCounterOverflow,
            signed(Value_ob) => PwmCounter_b
        );
    
    pPwmValueReg: process (Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if PwmCounterOverflow = '1' then
                PwmValue_bd <= PwmValue_b;
            end if;
        end if;
    end process;

    pHearBeat: process (Clk_ik) begin
        if rising_edge(Clk_ik) then
            if Reset_ir = '1' then
                HeartBeat_o <= '0';
            else
                if PwmCounter_b = 0 then
                    HeartBeat_o <= '1';
                end if;
                if PwmCounter_b = PwmValue_bd then
                    HeartBeat_o <= '0';
                end if;
            end if;
        end if;
    end process;

end architecture;
